"""
This module contains the extension's core functionality. 
"""

from __future__ import annotations

from functools import wraps
from logging import getLogger
from typing import Any, Callable, Dict, Mapping, TYPE_CHECKING

from flask import Flask, redirect, current_app, g, request, session
from ipernity import IpernityAPI
from werkzeug.local import LocalProxy

if TYPE_CHECKING:
    from flask import Response
    from flask_login import LoginManager


log = getLogger(__name__)


default_flask_options = {
    'IPERNITY_CACHE_REQUESTS': False,
    'IPERNITY_CACHE_MAX_AGE': 300,
    'IPERNITY_CALLBACK': True,
    'IPERNITY_CALLBACK_URL_PREFIX': '/ipernity',
    'IPERNITY_LOGIN': False,
    'IPERNITY_LOGIN_URL_PREFIX': '/ipernity',
    'IPERNITY_PERMISSIONS': {},
    'IPERNITY_SESSION_PREFIX': 'ipernity_',
}


class Ipernity():
    """
    
    Args:
        app:            The Flask application
    """
    
    def __init__(
        self,
        app: Flask|None = None,
    ):
        # Initialize app
        if app is not None:
            self.init_app(app)
    
    
    def init_app(self, app: Flask):
        log.debug('Initializing Ipernity with app %s', app.name)
        app.extensions['ipernity'] = self
        
        # Default configuration
        for option, value in default_flask_options.items():
            app.config.setdefault(option, value)
        
        if app.config['IPERNITY_CALLBACK']:
            log.debug('Preparing callback blueprint')
            from .callback import callback
            app.register_blueprint(
                callback,
                url_prefix = app.config['IPERNITY_CALLBACK_URL_PREFIX']
            )
        
        if app.config['IPERNITY_LOGIN']:
            log.debug('Preparing login manager blueprint')
            from .login import ip_login, init_app as init_login
            app.register_blueprint(
                ip_login,
                url_prefix = app.config['IPERNITY_LOGIN_URL_PREFIX']
            )
            init_login(app)
        
        # 
        app.context_processor(_context_processor)
    
    
    def authorize(
        self,
        permissions: Mapping|None = None,
        next_url: str|None = None
    ) -> Response:
        if permissions is None:
            permissions = current_app.config['IPERNITY_PERMISSIONS']
        log.debug('Authorizing for %s', permissions)
        if next_url is None:
            next_url = request.url
        self.session_set('next_url', next_url)
        return redirect(self.api.auth.auth_url(permissions))
    
    
    def set_token(self, frob: str|None = None):
        log.debug('Setting token')
        if frob is None:
            log.debug('Getting frob from request')
            frob = request.args.get('frob')
        log.debug('Got frob %s', frob)
        
        # Get token and save it to session and API
        self.session_set('token', self.api.auth.getToken(frob)['auth'])
    
    
    def logout(self):
        for key in list(session):
            if key.startswith(current_app.config['IPERNITY_SESSION_PREFIX']):
                del session[key]
        self.api.token = None

    
    @property
    def api(self) -> IpernityAPI:
        """The current Ipernity API"""
        if 'ipernity_api' not in g:
            log.debug('Creating IpernityAPI object')
            kwargs = {
                'api_key':      current_app.config['IPERNITY_APP_KEY'],
                'api_secret':   current_app.config['IPERNITY_APP_SECRET'],
                'token':        self.session_get('token'),
                'auth':         'web',
            }

            if current_app.config['IPERNITY_CACHE_REQUESTS']:
                from .cache import CachedIpernityAPI
                g.ipernity_api = CachedIpernityAPI(
                    current_app.config['IPERNITY_CACHE_MAX_AGE'],
                    **kwargs
                )
            else:
                g.ipernity_api = IpernityAPI(**kwargs)
        
        return g.ipernity_api
    
    
    def session_get(self, key: str, default: Any = None) -> Any:
        """
        Returns a session variable.
        """
        return session.get(current_app.config['IPERNITY_SESSION_PREFIX'] + key, default)
    
    
    def session_set(self, key: str, value: Any):
        """Sets a session variable."""
        session[current_app.config['IPERNITY_SESSION_PREFIX'] + key] = value
    
    
    def session_pop(self, key: str, default: Any = None) -> Any:
        """
        Removes a session variable and returns it.
        """
        return session.pop(current_app.config['IPERNITY_SESSION_PREFIX'] + key, default)


def _context_processor() -> Dict:
    return {
        'ipernity': ipernity,
    }


def _get_ipernity() -> Ipernity:
    return current_app.extensions['ipernity']


ipernity: Ipernity = LocalProxy(_get_ipernity)


def ipernity_auth_required(permissions: Mapping[str,str]|None = None) -> Callable:
    """
    Decorator for a view that requires Ipernity authentication.
    """
    def decorate(f: Callable) -> Callable:
        nonlocal permissions
        @wraps(f)
        def view(*args, **kwargs):
            nonlocal permissions
            if permissions is None:
                permissions = current_app.config['IPERNITY_PERMISSIONS']
            if (
                ipernity.api.token is None or
                not ipernity.api.has_permissions(permissions)
            ):
                return ipernity.authorize(permissions)
            return f(*args, **kwargs)
        
        return view
    
    return decorate






from __future__ import annotations

import os
from html.parser import HTMLParser
from logging import getLogger
from urllib.parse import parse_qs, urlparse
from typing import Dict, Mapping, TYPE_CHECKING

import pytest
import requests
import yaml
from flask import Flask, jsonify, session
from flask_ipernity import Ipernity, ipernity

if TYPE_CHECKING:
    from flask.testing import FlaskClient


log = getLogger(__name__)


@pytest.fixture(scope='session')
def test_config() -> Dict:
    cfgfile = os.path.join(os.path.dirname(__file__), '.test-config.yaml')
    with open(cfgfile, 'r') as cfg:
        return yaml.load(cfg, Loader=yaml.SafeLoader)


@pytest.fixture
def base_app(test_config: Mapping) -> Flask:
    a = Flask(__name__)
    a.config.from_mapping(test_config['flask'])
    a.testing = True
    
    @a.route('/get_token')
    def get_token():
        log.debug('Returning Ipernity token')
        return jsonify(ipernity.session_get('token'))
    
    @a.route('/get_perm')
    def get_perm():
        log.debug('Returning Ipernity permissions')
        return jsonify(ipernity.api.permissions)
    
    return a


@pytest.fixture
def browser(test_config: Mapping) -> IpernitySession:
    br = IpernitySession()
    for domain, paths in test_config['ipernity']['cookies'].items():
        for path, cookies in paths.items():
            for name, value in cookies.items():
                br.cookies.set(name, value, domain=domain, path=path)
    return br


class IpernitySession(requests.Session):
    def authorize(self, auth_url: str) -> str:
        log.info('Authorizing via %s', auth_url)
        res = self.get(auth_url, allow_redirects=False)
        mimetype = res.headers.get('Content-Type')
        log.info('Got result %d: %s', res.status_code, mimetype)
        
        if res.status_code == 200 and mimetype.startswith('text/html'):
            log.debug('Text: %s', res.text)
            html = IpernityParser()
            html.feed(res.text)
            url = urlparse(auth_url)
            log.info('Posting authorization %s', html.params)
            res = self.request(
                html.method,
                auth_url,
                data = html.params,
                allow_redirects = False
            )
            mimetype = res.headers.get('Content-Type')
            log.info('Got result %d: %s', res.status_code, mimetype)
        
        if res.status_code == 302:
            url = res.headers.get('location')
            log.info('Got redirect to %s', url)
            if url.startswith('http://127.0.0.1'):
                frob = parse_qs(urlparse(url).query)['frob'][0]
                log.info('Returning frob %s', frob)
                return frob
        
        if mimetype.startswith('text/html'):
            log.debug('Text: %s', res.text)
        
        log.error('No frob')
        return None


class IpernityParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.parsing_form = False
        self.params = {}
    
    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]):
        if tag == 'form':
            if self.get_attr(attrs, 'name') == 'fa':
                self.method = self.get_attr(attrs, 'method')
                self.parsing_form = True
        
        elif (
            self.parsing_form and
            tag == 'input' and
            self.get_attr(attrs, 'type') == 'hidden'
        ):
            self.params[self.get_attr(attrs, 'name')] = self.get_attr(attrs, 'value')
    
    def handle_endtag(self, tag: str):
        if tag == 'form' and self.parsing_form:
            self.parsing_form = False
    
    @staticmethod
    def get_attr(attrs: list, key: str) -> str|None:
        for k, v in attrs:
            if k == key:
                return v
        return None







from __future__ import annotations

from logging import getLogger
from typing import Any, TYPE_CHECKING

from flask_ipernity import Ipernity
import pytest

if TYPE_CHECKING:
    from flask import Flask
    from flask.testing import FlaskClient


log = getLogger(__name__)


@pytest.fixture
def app(base_app: Flask) -> Flask:
    from flask import jsonify, url_for, redirect
    from flask_login import LoginManager, login_required, current_user
    
    a = base_app
    a.config['IPERNITY_LOGIN'] = True
    LoginManager(a)
    Ipernity(a)
    
    @a.route('/user')
    def user():
        log.debug('Returning user info')
        return jsonify({
            'id':               current_user.get_id(),
            'is_active':        current_user.is_active,
            'is_anonymous':     current_user.is_anonymous,
            'is_authenticated': current_user.is_authenticated,
        })
    
    @a.route('/redir')
    def redir():
        return redirect(url_for(
            'ip_login.login'
        ))
    
    @a.route('/restricted')
    @login_required
    def restricted():
        return 'restricted'
    
    return a


@pytest.fixture
def client(app: Flask) -> FlaskClient:
    return app.test_client()


def test_login_logout(browser, client, test_config):
    res = client.get('/ipernity/login')
    frob = browser.authorize(res.location)
    assert frob
    log.info('Calling callback with frob %s', frob)
    res = client.get('/ipernity/cb', query_string = {'frob': frob})
    log.info('Got result %d: %s', res.status_code, res.headers.get('content-type'))
    log.info('Checking token')
    res = client.get('/get_token')
    log.info(
        'Got result %d: %s (%s)',
        res.status_code,
        res.headers.get('content-type'),
        res.text
    )
    assert res.json['user']['username'] == test_config['ipernity']['username']
    log.info('Checking authentication status')
    res = client.get('/user')
    assert res.json['is_authenticated']
    log.info('Logging out')
    res = client.get('/ipernity/logout')
    log.info('Got result %d: %s', res.status_code, res.headers.get('content-type'))
    res = client.get('/get_token')
    assert res.json is None


def test_not_logged_in(client):
    log.info('Checking user status')
    res = client.get('/user')
    assert not res.json['is_active']
    assert not res.json['is_authenticated']
    assert res.json['is_anonymous']


def test_login_required(app, client):
    log.info('Getting login URL')
    res1 = client.get('/redir')
    log.info('Checking redirect of restricted view')
    res2 = client.get('/restricted')
    assert res2.location.startswith(res1.location + '?')




"""
Tests cache functionality
"""

from time import sleep

from flask import jsonify
import pytest

from flask_ipernity import Ipernity, ipernity


@pytest.fixture
def cached_app(base_app):
    app = base_app
    app.config['IPERNITY_CACHE_REQUESTS'] = True
    
    Ipernity(app)
    
    @app.route('/cache')
    def cache():
        return jsonify({
            key: ipernity.session_get(key, 0)
            for key in ['api_calls', 'returns_from_cache']
        })
    
    @app.route('/explore')
    def explore():
        return jsonify(ipernity.api.explore.docs.getPopular())
    
    return app


def test_cache(cached_app):
    cached_app.config['IPERNITY_CACHE_MAX_AGE'] = 900
    client = cached_app.test_client()
    client.get('/explore')
    res = client.get('/cache')
    api_calls = res.json['api_calls']
    cached_calls = res.json['returns_from_cache']
    client.get('/explore')
    res = client.get('/cache')
    assert res.json['api_calls'] == api_calls
    assert res.json['returns_from_cache'] > cached_calls


def test_cache_expire(cached_app):
    cached_app.config['IPERNITY_CACHE_MAX_AGE'] = 1
    client = cached_app.test_client()
    client.get('/explore')
    res = client.get('/cache')
    api_calls = res.json['api_calls']
    cached_calls = res.json['returns_from_cache']
    sleep(1)
    client.get('/explore')
    res = client.get('/cache')
    assert res.json['api_calls'] > api_calls
    assert res.json['returns_from_cache'] == cached_calls





"""Auth Blueprint"""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from flask import Blueprint, Response, redirect, current_app

from .ext import ipernity


log = getLogger(__name__)

callback = Blueprint('ip_callback', __name__)


@callback.route('/cb')
def ipernity_callback() -> Response:
    log.debug('Ipernity callback called')
    ipernity.set_token()
    if current_app.config['IPERNITY_LOGIN']:
        from .login import do_login
        do_login()
    next_url = ipernity.session_pop('next_url', '/')
    return redirect(next_url)





"""
Tests cache functionality
"""

from __future__ import annotations

from logging import getLogger
from time import sleep
from typing import TYPE_CHECKING

from flask import Flask, jsonify
import pytest

from flask_ipernity import Ipernity, ipernity_auth_required, ipernity

if TYPE_CHECKING:
    from flask.testing import FlaskClient


log = getLogger(__name__)


@pytest.fixture
def app(base_app):
    app = base_app
    Ipernity(app)
    
    @app.route('/user')
    @ipernity_auth_required()
    def user():
        return jsonify(ipernity.api.user.get())
    
    @app.route('/albums')
    @ipernity_auth_required({'doc': 'read'})
    def albums():
        return jsonify(ipernity.api.albums.getList())
    
    return app


@pytest.fixture
def client(app: Flask) -> FlaskClient:
    return app.test_client()


def test_auth_required(client, browser, test_config):
    res = client.get('/user')
    assert res.location
    frob = browser.authorize(res.location)
    assert frob
    res = client.get('/ipernity/cb', query_string = {'frob': frob})
    assert res.location.endswith('/user')
    res = client.get('/user')
    assert res.json['user']['username'] == test_config['ipernity']['username']
    res = client.get('/get_perm')
    assert res.json['doc'] == 'none'
    res = client.get('/albums')
    assert res.location
    frob = browser.authorize(res.location)
    assert frob
    res = client.get('/ipernity/cb', query_string = {'frob': frob})
    assert res.location.endswith('/albums')
    res = client.get('/get_perm')
    assert res.json['doc'] == 'read'





"""
This test must be run before any other test that uses Flask-Login.
Be careful when renaming files.
"""

import sys
from logging import getLogger

from flask_ipernity import Ipernity, ipernity

log = getLogger(__name__)


def test_no_login_no_cache(base_app, browser, test_config):
    app = base_app
    Ipernity(app)
    
    @app.route('/login')
    def login():
        return ipernity.authorize({})
    
    client = app.test_client()
    res = client.get('/login')
    frob = browser.authorize(res.location)
    assert frob
    log.info('Calling callback with frob %s', frob)
    res = client.get('/ipernity/cb', query_string = {'frob': frob})
    log.info('Got result %d: %s', res.status_code, res.headers.get('content-type'))
    log.info('Checking token')
    res = client.get('/get_token')
    log.info(
        'Got result %d: %s (%s)',
        res.status_code,
        res.headers.get('content-type'),
        res.text
    )
    assert res.json['user']['username'] == test_config['ipernity']['username']
    assert 'flask_login' not in sys.modules
    assert 'flask_ipernity.login' not in sys.modules
    assert 'flask_ipernity.cache' not in sys.modules







"""
"""

from __future__ import annotations

from logging import getLogger
from time import time
from typing import Any, Dict, Mapping, TYPE_CHECKING

from ipernity import IpernityAPI

from .ext import ipernity

# if TYPE_CHECKING:


log = getLogger(__name__)


class CachedIpernityAPI(IpernityAPI):
    """
    Wrapper for :class:`~ipernity.IpernityAPI` that caches requests
    in the Flask Session.
    
    Args:
        timeout:    Time in seconds that cached results are considered valid.
        args:       Passed to :class:`~ipernity.IpernityAPI`.
        kwargs:     Passed to :class:`~ipernity.IpernityAPI`.
    """
    
    def __init__(
        self,
        timeout: int = 300,
        *args: Any,
        **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self.timeout = timeout
    
    
    @property
    def cache(self) -> Dict:
        """Dict for storing the cache"""
        cache = ipernity.session_get('cache', None)
        if cache is None:
            log.debug('Initializing cache')
            ipernity.session_set('cache', {})
            cache = ipernity.session_get('cache', None)
        return cache
    
    
    def call(self, method_name: str, **kwargs: Any) -> Mapping:
        key = method_name + repr(self.token) + repr(kwargs)
        if key in self.cache:
            res, expire = self.cache[key]
            if time() < expire:
                log.debug(
                    '%s(%s): returning result from cache',
                    method_name,
                    kwargs
                )
                ipernity.session_set(
                    'returns_from_cache',
                    ipernity.session_get('returns_from_cache', 0) + 1
                )
                return res
        
        res = super().call(method_name, **kwargs)
        
        # This will also set session.modified
        ipernity.session_set(
            'api_calls',
            ipernity.session_get('api_calls', 0) + 1
        )
        
        self.cache[key] = (res, time() + self.timeout)
        return res





"""
Authentication Flask Extensions
"""

from __future__ import annotations

from logging import getLogger
from typing import List, Mapping, TYPE_CHECKING

from flask import (
    Blueprint, Flask, Response,
    redirect,
    current_app, request
)
from flask_login import (
    UserMixin,
    login_user, logout_user,
    user_logged_out
)
from werkzeug.local import LocalProxy

from .ext import ipernity

if TYPE_CHECKING:
    from ipernity import IpernityAPI
    from .ext import Ipernity


log = getLogger(__name__)    


ip_login = Blueprint('ip_login', __name__)


def init_app(app: Flask):
    app.login_manager.login_view = 'ip_login.login'
    app.login_manager.user_loader(load_user)
    user_logged_out.connect(on_logout)


@ip_login.route('/login')
def login() -> Response:
    log.debug('Login called')
    return ipernity.authorize(
        current_app.config['IPERNITY_PERMISSIONS'],
        request.args.get('next')
    )


@ip_login.route('/logout')
def logout() -> Response:
    """Logout"""
    log.debug('Logout called')
    logout_user()
    return redirect('/')


def do_login():
    login_user(User(ipernity.api))


def load_user(id_: str) -> User:
    """
    Loads a user from the session.
    """
    log.debug('Loading user %s', id_)
    if not ipernity.api.token:
        return None
    if not ipernity.api.user_info:
        log.error('No user info for API token, dumping token')
        ipernity.logout()
        return None
    if ipernity.api.user_info['user_id'] != id_:
        log.error(
            'User mismatch (given %s, token %s), dumping token',
            id_,
            ipernity.api.user_info['user_id']
        )
        ipernity.logout()
        return None
    
    return User(ipernity.api)


def on_logout(sender: Flask, user: User):
    ipernity.logout()


class User(UserMixin):
    """
    A user
    """
        
    def __init__(
        self,
        api: IpernityAPI
    ):
        log.debug('Creating Ipernity user')
        self._api = api
    
    def __repr__(self) -> str:
        return '<User({})>'.format(self.get_id())
    
    def get_id(self) -> str:
        return self._api.user_info['user_id']
    
    @property
    def is_authenticated(self) -> bool:
        if self._api.token is None:
            return False
        return True
    
    @property
    def username(self) -> str:
        if self.is_authenticated:
            return self._api.user_info['username']
        else:
            return ''
    
    @property
    def realname(self) -> str:
        if self.is_authenticated:
            return self._api.user_info['realname']
        else:
            return ''


